/**************************************************************************
 *
 * linux_lvnode.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * This is the public interface to the linux-specific lvnode data
 * type.  The lvnode data type provides an abstract non-OS-dependent
 * interface to the linux-specific inode data type.  LOMAC uses the
 * lvnode data type to shield the majority of its code from
 * linux-specific inode details.
 *
 **************************************************************************/

#include <linux/net.h>            /* for socket and protocol family structs */
#include <linux/net_alias.h>      /* for struct net_alias */
#include <linux/netdevice.h>      /* for struct device    */
#include <net/sock.h>             /* for struct sock      */
#include <linux/skbuff.h>         /* for struct sk_buff   */


#include "kernel_lgid.h"
#include "kernel_lpid.h"
#include "kernel_lvnode.h"
#include "kernel_lnic.h"

#include "lomac_nlm.h"

/* const */
lvnode_t *lvnode_from_path( const char *path_s ) {
  lvnode_t *p_vnode;
  if( open_namei( path_s, 0x00, 0x00, &p_vnode, NULL ) ) {
    /* if( namei( path_s, &p_vnode ) ) { */
    return( NULL );
  } else {
    iput( p_vnode );
    return( p_vnode );
  }
}

/* network_socket_to_level()
 *
 * in:     p_socket - pointer to the socket whose current level we wish
 *                    to determine.
 * out:    p_level  - set to the level of the lowest-levelled NIC which
 *                    has a sk_buff enqueued on socket `p_socket', or
 *                    0 if no sk_buffs or error.
 *         p_name_s - set to the name of the lowest-levelled NIC which
 *                    has a sk_buff enqueued on socket `p_socket', or
 *                    NULL if no sk_buffs or error.
 * return: value      condition
 *         -----      ---------
 *           1        There were sk_buffs enqueued on `p_socket'
 *           0        There were no sk_buffs enqueued on `p_socket'
 *
 * On a multi-homed host (a host with more than one NIC), at any given
 * time, a single network socket (that is, a non-PF_UNIX socket) can
 * have sk_buffs originating from one or more of its NICs enqueued.
 * Each NIC can have a different level according to LOMAC's policy
 * input.  So, at any given time, the level of a network socket is the
 * minimum of the levels of all the NICs which have sk_buffs enqueued.
 *
 * This function traverses the queue of sk_buffs.  If the queue is
 * non-empty, it sets `p_level' to the level of the socket.  It also
 * saves a pointer to the name of the NIC with the lowest level in
 * `p_name_s'.  If the queue is empty, the read will return no data,
 * and may be safely ignored by LOMAC.  In this case, it returns 0.
 *
 * An open question - if a read system call will return data from only
 * the first sk_buff in the queue, then we really only want to examine
 * the level of this first sk_buff, instead of taking the minimum of
 * the entire queue.
 *  */

int lvnode_network_socket_to_level( lvnode_t *p_lvnode, level_t *p_level,
				    char **p_name_s ) {

  struct socket *p_socket;      /* points to `p_lvnode's socket struct */
  struct sk_buff_head p_rq;     /* points to `p_socket's receive queue */
  unsigned int skbuffs_left;    /* counts enqueued skbuffs */
  struct sk_buff *p_sk_buff;    /* pointer to enqueued buffers of net data */
  struct device *p_nic;         /* pointer to device structs for NICs */
  level_t level = 0;            /* level of current skbuff */

  /* Clear `p_level' and `p_name_s' in case the queue is empty. */
  (*p_level)  = 0;
  (*p_name_s) = NULL;           

  /* Make `p_socket' point to the socket structure *
   * encapsulated in `p_lvnode'.                   */
  p_socket = &( p_lvnode->u.socket_i );

  /* We want to browse through the sk_buffs enqueued for `p_socket'. *
   * Make `p_rq' a pointer to `p_socket's receive queue.             */
  /* Is struct sock unique to sockets using the TCP protocol? *
   * Will this break with other protocols?                    */
  p_rq = ((struct sock *)p_socket->data)->receive_queue;

  /* We will now iterate through the queue of received skbuffs.  We
   * will limit our for loop with `p_rq.qlen' rather than watching for
   * a NULL `next' pointer because experience has shown that the next
   * pointer of the last skbuff enqueued is, in fact, not NULL.
   * Hmmm... this didn't happen in FreeBSD.                            */

  if( !( skbuffs_left = p_rq.qlen ) ) {

    /* Experience has shown that the act of sending a single keystroke
     * through an established connection (perhaps when sending your
     * password through rlogin) generates at least one sk_buff on the
     * receive queue.  However, the r_rq.qlen count will remain zero.
     * The first (and perhaps only) sk_buff on the queue will have a
     * non-null dev (NIC) pointer, but the pointer will not point to
     * any of the existing NIC device structures.  Although the
     * struct's name pointer is non-null, dereferencing it leads to a
     * general protection fault, suggesting that it is not pointing to
     * a valid device structure at all.  Hmm... this didn't happen in
     * FreeBSD.                                                        */

    return( 0 );
  }

  /* `p_rq->next' is the sk_buff at the head of `p_socket's receive *
   * queue.  Is it sufficient to examine it, or must we iterate     *
   * through the entire queue and examine all packets?  I don't     *
   * know - we'll iterate for now.                                  */

  for( p_sk_buff = p_rq.next; skbuffs_left; p_sk_buff = p_sk_buff->next ) {

    p_nic = p_sk_buff->dev;
    skbuffs_left--;         /* one more skbuff done, one less to go */
    level = nic_to_level( p_nic );
    if( ( !(*p_level) ) || ( level < (*p_level) ) ) {
      (*p_level)  = level;            /* save the level of the NIC */
      (*p_name_s) = p_nic->name;      /* save the name of the NIC */
    }

  } /* for all enqueued skbuffs */
    
  return( 1 );  /* The queue was not empty, return 1 */

} /* lvnode_network_socket_to_level() */


