/************************************************************************
 *
 * linux_driver.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 * Routines to manage LOMAC's user-space interface: the /dev/lomac
 * pseudodevice.  User-space utility programs use this device to
 * set the LOMAC policy, query the state of the LOMAC LKM, and
 * determine the levels of processes and files, etc. during run-time.
 *
 *
 * References:
 *
 * McKusick et. al. `The Design and Implementation of the 4.4 BSD
 * Operating System,' Addison-Wesley 1996, section 6.3 Character
 * Devices.
 *
 * Allesandro Rubini `Dynamic Kernels: Modularized Device Drivers' in
 * Linux Journal #23, March 1996.
 *
 * Allesandro Rubini `Dynamic Kernels: Discovery' in Linux Journal #24,
 * April 1996.
 *
 *
 ************************************************************************/


#include "kernel_kernel.h"   /* for memory allocation stuff */
#include "kernel_lgid.h"     /* lvnode.h needs lgid_t */
#include "kernel_lpid.h"     /* lomac_usi.h needs lpid_t */
#include "kernel_lvnode.h"   /* to include <linux/fs.h> indirectly */

#include <asm/uaccess.h>    /* for verify_area() */

#include "lomac_level.h"
#include "lomac_usi.h"

/* Commands for ioctl() interface. */
#define GET_PID_LEVEL_IOCTL_CMD  0
#define GET_FILE_LEVEL_IOCTL_CMD 1


int do_ioctl( struct inode *, struct file *, unsigned int, unsigned long );

/* This is the file operation jump table for the LOMAC *
 * character device.  We'll be using ioctl, only.      */
struct file_operations lomac_fops = { NULL, NULL, NULL, NULL, NULL,
				      do_ioctl, NULL, NULL, NULL, NULL,
				      NULL, NULL, NULL };


/* ----------------------- Private Functions ----------------------- */


/* do_ioctl()
 *
 * in:     p_inode_ignored - ignored
 *         p_file_ignored  - ignored
 *         command - GET_PID_LEVEL_IOCTL_CMD  - return level of pid in `arg'.
 *                   GET_FILE_LEVEL_IOCTL_CMD - return level of file named
 *                                              by the string `arg'.
 *         arg     - if `command' is GET_PID_LEVEL_IOCTL_CMD, then `arg' is
 *                   the pid of the process whose level we want to query.
 *                   if `command' is GET_FILE_LEVEL_IOCTL_CMD, then `cmd'
 *                   points to the name of the file whose level we want
 *                   to query, stored as a null-terminated string.
 * out:    nothing
 * return: the level of the process or file indicated by `command' and `arg',
 *         or -1 if no such process or file exists.  Also returns -1 on
 *         a bad `cmd' parameter.
 *
 *     User-space utility programs may query the state of the LOMAC LKM
 * through this ioctl() interface.
 *
 */

int do_ioctl( struct inode *p_inode_ignored, struct file *p_file_ignored,
	      unsigned int command, unsigned long arg ) {

  level_t level;       /* holds level of process or file indicated by `arg' */
  unsigned int filename_length;       /* length of file name from userspace */
  char *filename_s;             /* buffer to hold filenames from user-space */

  switch( command ) {
  case GET_PID_LEVEL_IOCTL_CMD:          /* look up process level */

    level = usi_get_pid_level( arg );
    break;

  case GET_FILE_LEVEL_IOCTL_CMD:         /* look up file level */

    /* When a file level is queried, we expect `arg' to point to a     *
     * variable-sized block of user-space memory organized as follows: *
     * First, an unsigned int containing the size of a following       *
     * buffer, and second, the buffer itself.  The buffer is used to   *
     * contain the absolute path to the file in question, as a         *
     * null-terminated string.                                         */

    /* First, verify that we can read the string-length uint, *
     * and copy it into kernel memory as `filename_length'.   */
    if( verify_area( VERIFY_READ, (void *)arg, sizeof( unsigned int ) ) ) {
      return( -1 );
    }
    copy_from_user( &filename_length, (void *)arg, sizeof( int ) );

    /* Second, make sure that `filename_length' is a proper filename *
     * length, verify that we can read the string in user-space,     *
     * allocate a kernel-space buffer for it, and copy the filename  *
     * string into the kernel-space buffer. We're assuming void * is *
     * equivalent to char * here.                                    */

    if( ( filename_length <= 0 ) || ( filename_length > PATH_MAX ) ) {
      return( -1 );
    }

    arg = (unsigned int)((unsigned int *)arg + 1);  /* skip over size uint */
    if( verify_area( VERIFY_READ, (void *)arg, filename_length ) ) {
      return( -1 );
    }

    if( !( filename_s = (char *)lmalloc( filename_length ) ) ) {
      lpanic( "LOMAC (lomac_usi.c) : Unable to allocate filename buffer" );
    }

    copy_from_user( filename_s, (void *)arg, filename_length );

    /* ensure NULL termination */
    filename_s[ filename_length - 1 ] = '\0';

    /* verify that `filename_s' is an absolute path */
    if( filename_s[ 0 ] != '/' ) {
      return( -1 );
    }

    level = usi_get_file_level( filename_s );

    lfree( filename_s );
    break;
    
  default:                               /* error - bad cmd value */

    return( -1 );

  } /* switch( command ) */

  return( level );

} /* do_ioctl() */


/* ------------------------ Public Functions ----------------------- */


/* init_driver()
 *
 * in:     nothing
 * out:    nothing
 * return: nothing
 *
 */

void init_driver( void ) {

  /* Register a character device for LOMAC.  The 0 in the first parm  *
   * indicates that we want the kernel to dynamically allocate a free *
   * device major number for us.  User-space programs will be able to *
   * determine our device major number by examining the file          *
   * /proc/devices.  This will fail only if there are no free major   *
   * device numbers left for the kernel to allocate.                  */

  if( !register_chrdev( 0, "lomac", &lomac_fops ) ) {
    lpanic( "LOMAC(lomac_usi.c) no major device number available for lomac" );
  }

} /* init_driver() */
