/***********************************************************************
 *
 * lomac_fsolm.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 * This file contains the implementation of the FileSystem Object Level
 * Map (FSOLM).  
 *
 * The FSOLM is a mapping of filesystem objects to levels.  As far as
 * the FSOLM is concerned, filesystem objects include any object that
 * is represented by an inode: files (directories, etc), unnamed
 * pipes, named pipes (FIFOs), and local communication sockets.  The
 * following text describes the requirements of the FSOLM, and how it
 * uses three special data structures (the GLM, ELM, and VLM) to meet
 * those requirements.
 *
 * The FSOLM is a means of associating security attributes (in this
 * case, levels) with filesystem objects.  The standard Linux kernel
 * already associates some security attrbutes, such as owner and group
 * information, with filesystem objects.  The kernel stores these
 * security attributes in the filesystem inodes.  Storing the security
 * attributes in the inodes is a good solution, because: (1) since you
 * always have the inode handy when dealing with a file, the security
 * attributes can be found very efficiently, and (2) inodes are backed
 * up on disk, making the security attribute information persistent
 * across reboots.  Unfortunately, there is no room to store LOMAC
 * levels in the existing inode structures.  So, we must maintain some
 * new kind of data structure to associate levels with file system
 * objects that (hopefully) is as efficient and persistent as the
 * existing inode-based scheme.
 *
 * The TIS Labs Domain and Type Enforcement (DTE) system solves this
 * problem with a data structure called the Hierarchical Attribute
 * DataBase (HADB).  The HADB is both efficient and (on some versions
 * of the DTE prototype) persistent.  However, it is not trivial to
 * implement, and works best with the help of a small set of extra
 * function calls inserted into various parts of the namei() name
 * resolution mechanism.  Providing these extra function calls with
 * LOMAC would require duplicating the kernel's namei() mechanism in
 * the LKM, or replacing it via binary patching.  Rather than employ
 * such a heavyweight solution, I have taken Lee Badger's suggestion
 * to rely on a simpler approach.  This approach is the FSOLM, and is
 * a hack.  Those interested in learning how to solve the security
 * attribute mapping problem in an elegant fashion might consider
 * studying HADB rather than FSOLM.
 * 
 * Note:  binary patching is no longer taboo for LOMAC.  However,
 *        since HADB is US Patent 5,819,275, I'll stick with FSOLM
 *        in order to keep LOMAC Free. 
 *
 * In order to achieve efficiency and persistence cheaply, the FSOLM
 * overloads the existing inode-based security attribute storage
 * scheme to maintain most of its own filesystem object level
 * mappings.  Specifically, it maps inode gid attributes to levels.
 * For example, any file whose inode's gid is "root", "sys", or "disk"
 * is level 3, "users" corresponds to level 2, and all others are 1.
 * The FSOLM maintains this mapping with a data structure called the
 * Group Level Map (GLM).  This overloading makes LOMAC level
 * attributes (almost) as efficient to examine and as persitent as the
 * existing group id attributes.
 *
 * The group-id overloading technique is a *good thing* provided that
 * in the vast majority of cases, the existing gid attribute of a file
 * is (coincidentally) set to a group that is suitable both for the
 * standard Linux access control policy and for the LOMAC access
 * control policy.  With the traditional Linux access control policy's
 * group arrangement, this appears to be a reasonably good assumption.
 * Still, convenient coincidences will not occur in every case.  The
 * socket "/dev/log", for instance, is traditionally set to gid "root"
 * on Linux systems.  However, most LOMAC policies will want to set
 * "/dev/log" to the lowest possible level, in order to allow all
 * processes to write log messages into it.  We cannot map the "root"
 * group to the lowest level, because the vast majority of the other
 * "root"-group files deserve the protection of a higher level.  One
 * solution to this problem would be to use chgrp to change the gid of
 * "/dev/log" to some other group, such as "user".  Unfortunately,
 * this solution is entirely unnacceptable, for two reasons: First, we
 * do not wish to require users to modify their existing
 * configurations in any way in order to try LOMAC.  Second, such
 * group changes could break the protection provided by the
 * traditional Linux access control policy.
 *
 * In order to avoid the use of the unacceptable chgrp solution, the
 * FSOLM maintains an Exception Level Map (ELM).  The ELM is a list of
 * exceptions to the GLM's rule.  These exceptions take the form of
 * <pathname,level> pairs, where the level of the file named by
 * pathname is set to the level indicated by the pair, regardless of
 * the file's gid.  Conceptually, this leads to the following
 * algorithm to determine the level of a given file "file":
 *
 * IF "file" has an entry in the ELMC:
 *   THEN report the level of "file" as indicated by the ELM.
 *   ELSE report the level of "file" as indicated by the GLM.
 *
 * Since this level determination must be done on every read and
 * write, and checking the ELM requires slow string comparisons, in
 * practice the above algorithm would be much too slow.  Instead,
 * LOMAC maintains a third data structure called the Vnode Level Map
 * (VLM).  The VLM maps <inode number, device number, version number>
 * triplets to levels, making it much faster to search than the ELM.
 * The real level determination algorithm checks the VLM rather than
 * the ELM.  LOMAC performs all sorts of semi-crufty checks on several
 * not-so-frequently-called system calls to make sure the VLM agrees
 * with the ELM at all times.  More information on the ELM and VLM can
 * be found in lomac_vlm.c .
 *
 *
 *
 * References:
 *
 * Badger et. al., "A Domain and Type Enforcement UNIX Prototype,"
 * 5th USENIX Security Symposium, 1995.  Reprinted in USENIX
 * Computing Systems Volume 9, Winter 1996.  Good reference for HADB.
 *
 * The following two papers contain DTE policies which illustrate the 
 * file access rights several common daemons need in order to operate
 * correctly.  Some of these files are likely exceptions for the IELM
 * in a LOMAC policy.
 *
 * Fraser, T. and Badger, L., "Ensuring Continuity During Dynamic
 * Security Policy Recofiguration in DTE," 1998 IEEE Symposium on
 * Security and Privacy.  See figure 1 in particular.
 *
 * Walker et. al., "Confining Root Programs with Domain and Type
 * Enforcement," 6th USENIX Security Symposium, 1996.
 *
 * The following two books discuss the difficult and ultimately
 * hopeless task of attempting to do the job of mandatory access
 * control using the standard *NIX discretionary access control
 * mechanism.  They also describe the traditional arrangement of
 * groups that LOMAC should not disturb.
 *
 * Curry, David A., "UNIX System Security, A Guide for Users and
 * System Administrators," Chapter 3: File System Security,
 * Addison-Wesley, 1992.
 *
 * Frisch, AEleen, "Essential System Administration," Chapter 6:
 * Security - Protecting Files and the Filesystem, O'Reilly &
 * Associates, 1991-1995.
 *
 **************************************************************************/

#include "kernel_kernel.h"
#include "kernel_lgid.h"
#include "kernel_lpid.h"
#include "kernel_lvnode.h"
#include "kernel_lprocess.h"

#include "lomac_fsolm.h"
#include "lomac_plm.h"
#include "lomac_glm.h"
#include "lomac_vlm.h"
#include "lomac_elm.h"


/* -------------------------- Public Functions ---------------------- */


/* init_fsolm()
 * 
 * in:     nothing
 * out:    nothing
 * return: nothing
 *
 *
 */

void init_fsolm() {

  init_glm();
  init_elm();
  init_vlm();

} /* init_fsolm() */


/* fsolm_set_for_new_object()
 *
 * in:     p_lvnode        - lvnode whose level we want to set
 *         search_elm_flag - set this if `p_lvnode' has a name in the
 *                           file system, otherwise clear it.
 * out:    nothing
 * return: nothing
 *
 * This function updates the FSOLM to reflect the creation of a new
 * file system object.  LOMAC uses this function to assign a level to
 * each newly created object.  Once assigned, this level never
 * changes.  LOMAC uses the following algorithm to determine the level
 * of a new object:
 *
 * IF `search_elm_flag' is set
 *    IF `p_lvnode' corresponds to an entry in the ELM
 *       Add an entry to the VLM showing the level indicated by the ELM.
 *    ENDIF
 * ELSE 
 *    Add an entry to the VLM showing the level of the current process.
 * ENDIF
 *
 * The `search_elm_flag' parameter is a performance optimization.
 * Since the ELM specifies exceptions for file system objects by
 * filename, it can only contain exceptions for those objects that
 * have filenames.  If, when you call this function, you know that
 * `p_lvnode' does not have a filename (perhaps it is a socket or
 * pipe), then you can clear the `search_elm_flag' to avoid wasting
 * time searching the ELM.
 *
 * This function should be called only for the creation of objects
 * which did not exist, not the re-opening of existing objects.
 *
 */

void fsolm_set_for_new_object( lvnode_t *p_lvnode, int search_elm_flag ) {

  level_t current_process_level;   /* level of the current process */
  trust_t current_process_trust;   /* current process's trust setting */
  level_t elm_level;               /* level reported by the ELM */
  lvnode_id_t lvnode_id;           /* unique id for `p_lvnode' */

  /* Get the unique id for `p_lvnode' */
  lvnode_to_id( p_lvnode, &lvnode_id );

  /* If `search_elm_flag' is set, ask the ELM to assign *
   * a level to `p_lvnode'.  If it does, we are done.   */
  if( search_elm_flag  ) {

    if( ( elm_level =  elm_get( &lvnode_id ) ) ) {

      vlm_set( &lvnode_id, elm_level );

#ifdef PARANOID
      if( elm_level !=  fsolm_get( p_lvnode ) ) {
	lpanic( "LOMAC(lomac_fsolm.c) - "
	        "failed to update FSOLM on create via ELM" );
      }
#endif

      return;

    } /* if ELM handled the update */

  } /* if we should check the ELM */


  /* The ELM did not assign a level to `p_lvnode', so we must assign *
   * one ourselves.  The proper level of `p_lvnode' is the level of  *
   * the current process.  Start by finding the level of the current *
   * process.                                                        */

  plm_get( lprocess_pgrp( lprocess_current() ),
	   &current_process_level,
	   &current_process_trust );

#ifdef PARANOID
  if( ( current_process_level < LOMAC_LOWEST_LEVEL ) ||
      ( current_process_level > LOMAC_HIGHEST_LEVEL ) ) {
    lpanic( "LOMAC(lomac_fsolm.c) - bad process level on FSOLM update" );
  }
#endif

  /* Perhaps `p_inode's gid coincidentally maps to            *
   * `current_process_level' in the GLM.  If so, we are done. */
  if( gid_to_level( lvnode_gid( p_lvnode ) ) != current_process_level ) {

    /* `p_inode's gid attribute does not map to the proper level, so *
     * we must add an entry to the VLM indicating the proper level   *
     * ourselves.                                                    */
      
    vlm_set( &lvnode_id, current_process_level );

    /* Note that the FSOLM is not yet persistent across reboots.       *
     * Adding a VLM entry solves the mismatch between `p_lvnode's      *
     * proper level and the level indicated by its gid via the GLM for *
     * now.  As long as we never remove entries from the VLM, the      *
     * FSOLM will report the proper level for `p_lvnode' for the       *
     * remainder of the kernel's run-time.  However, if we reboot, the *
     * contents of the VLM will be lost.  If `p_lvnode' still exists   *
     * after the reboot, it will have the improper level indicated by  *
     * the GLM.                                                        */

  } else {

    /* Make sure there is no VLM entry hanging around for `lvnode_id' *
     * - one might have been left over by a file that was deleted or  *
     * clobbered by a rename.                                         */

    vlm_remove( &lvnode_id );

  } /* if VLM entry needed else VLM entry not needed */

#ifdef PARANOID

  if( fsolm_get( p_lvnode ) != current_process_level ) {

    lpanic( "LOMAC(lomac_fsolm.c) - failed to update VLM on create" );

  }

#endif PARANOID

} /* fsolm_set_for_new_object() */


/* fsolm_set_for_existing_object()
 *
 * in:     p_lvnode        - lvnode whose level we want to set
 *         search_elm_flag - set this if `p_lvnode' has a name in the
 *                           file system, otherwise clear it.
 * out:    nothing
 * return: nothing
 *
 * This function updates the FSOLM to reflect the re-opening of an
 * existing file system object.  LOMAC uses this function to load the
 * object's level assignment into the FSOLM, where it can be queried
 * efficiently.  LOMAC uses the following algorithm to determine the
 * level of an existing object:
 *
 * IF VLM already has a level entry for the object
 *    We're done, return.
 * ELSE IF `search_elm_flag' is set
 *    IF `p_lvnode' corresponds to an entry in the ELM
 *       return the level indicated by the ELM.
 *    ENDIF
 * ENDIF
 *
 * The `search_elm_flag' parameter is a performance optimization.
 * Since the ELM specifies exceptions for file system objects by
 * filename, it can only contain exceptions for those objects that
 * have filenames.  If, when you call this function, you know that
 * `p_lvnode' does not have a filename (perhaps it is a socket or
 * pipe), then you can clear the `search_elm_flag' to avoid wasting
 * time searching the ELM.
 *
 * This function should be called only for the re-opening of existing
 * objects, not for the creation of new objects.  
 *
 */

void fsolm_set_for_existing_object( lvnode_t *p_lvnode, int search_elm_flag ) {

  level_t elm_level;               /* level reported by the ELM */
  lvnode_id_t lvnode_id;           /* unique id for `p_lvnode' */

  /* Get the unique id for `p_lvnode' */
  lvnode_to_id( p_lvnode, &lvnode_id );

  /* Does `p_lvnode' have an entry in the VLM already?  If so, some    *
   * other process has already openend `p_lvnode', the VLM has already *
   * been made aware of it, and we are done.                           */
  if( vlm_get( &lvnode_id ) ) {

    return;

  }

  /* If `search_elm_flag' is set, ask the ELM to assign *
   * a level to `p_lvnode'.  If it does, we are done.   */
  if( search_elm_flag ) {

    if( ( elm_level = elm_get( &lvnode_id ) ) ) {

      vlm_set( &lvnode_id, elm_level );

#ifdef PARANOID
      if( elm_level != fsolm_get( p_lvnode ) ) {
	lpanic("LOMAC(lomac_fsolm.c) - reopen failed to update FSOLM via ELM");
      }
#endif

    } /* if ELM handled the update */

  } /* if we should search the ELM */


  /* The ELM did not assign a level to `p_lvnode', so we'll just live *
   * with the default specified by the GLM.                           */

#ifdef PARANOID
      if( ( fsolm_get( p_lvnode ) < LOMAC_LOWEST_LEVEL ) ||
	  ( fsolm_get( p_lvnode ) > LOMAC_HIGHEST_LEVEL ) ) {
	lpanic( "LOMAC(lomac_fsolm.c) - ELM failed to update FSOLM on reopen");
      }
#endif

} /* fsolm_set_for_existing_object() */


/* fsolm_set_to_same()
 *
 * in:     p_old_lvnode - lvnode with level set in FSLOM.
 *         p_new_lvnode - lvnode whose level we want to set to the
 *                        same level as `p_old_lvnode'.
 * out:    nothing
 * return: nothing
 *
 * This function adds an entry to the FSOLM for `p_new_lvnode', giving
 * it the same level as `p_old_lvnode'.
 *
 */

void fsolm_set_to_same( lvnode_t *p_old_lvnode, lvnode_t *p_new_lvnode ) {

  lvnode_id_t old_lvnode_id;       /* unique id for `p_old_lvnode' */
  lvnode_id_t new_lvnode_id;       /* unique id for `p_new_lvnode' */
  level_t level;                   /* level of `p_old_lvnode' */

  /* Get the unique ids for `p_old_lvnode' and `p_new_lvnode'. */
  lvnode_to_id( p_old_lvnode, &old_lvnode_id );
  lvnode_to_id( p_new_lvnode, &new_lvnode_id );

  /* Get the level of `p_old_lvnode'. */
  level = fsolm_get( p_old_lvnode );

#ifdef PARANOID
  if( !( level ) ) {
    lpanic( "LOMAC(lomac_fsolm.c) - bad level for dev %u ino %lu "
	    "found on set_to_same",
	    old_lvnode_id.device_number,
	    old_lvnode_id.lvnode_number );
  }
#endif

  /* If the level of `p_old_lvnode' is coincidentally the same as the *
   * level of `p_new_lvnode' according to the GLM, then we're done.   */
  if( level == gid_to_level( lvnode_gid( p_new_lvnode ) ) ) {
    return;
  }

  /* OK, we have to add an entry into the VLM for `p_new_lvnode' */
  vlm_set( &new_lvnode_id, level );

#ifdef PARANOID
      if( fsolm_get( p_old_lvnode ) != fsolm_get( p_new_lvnode ) ) {
	lpanic( "LOMAC(lomac_fsolm.c) - differing levels after set_to_same" );
      }
#endif

} /* fsolm_set_to_same() */



/* fsolm_get()
 *
 * in:     p_lvnode - lvnode whose level we want to know
 * out:    nothing
 * return: the level of `p_lvnode'
 *
 * Returns the level of the `p_lvnode'.  This task requires a
 * sequential check of the ILM and the GLM, as described above.
 *
 */

level_t fsolm_get( lvnode_t *p_lvnode ) {

  level_t level;             /* level of `p_lvnode' */
  lvnode_id_t lvnode_id;     /* unique id for `p_lvnode' */
  
  /* Get the unique id of `p_lvnode' */
  lvnode_to_id( p_lvnode, &lvnode_id );

  /* first, try to determine `p_lvnode_id's level via the ILM */
  if( ( level = vlm_get( &lvnode_id ) ) ) {
    return( level );
  }

  /* `p_lvnode_id' is not mentioned in the ILM; *
   *  determine its level via the GLM           */
  return( gid_to_level( lvnode_gid( p_lvnode ) ) );

} /* fsolm_get() */


/* will_confuse_glm()
 *
 * in:     old_gid
 *         new_gid
 * out:    nothing
 * return: val condition
 *         --- ---------
 *          0  `old_gid' and `new_gid' map to the same level in the GLM
 *          1  `old_gid' and `new_gid' don't map to the same level in the GLM
 */

int will_confuse_glm( lgid_t old_gid, lgid_t new_gid ) {

  return( gid_to_level( old_gid ) != gid_to_level( new_gid ) );

} /* will_confuse_glm() */




