/***************************************************************************
 *
 * lomac_plm.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 * The process->level map.
 *
 * The abstract LOMAC notion of a "subject" is equivalent to a process
 * group (also called a "job" in BSD) in the kernel.
 *
 * References:
 *
 * McKusick et. al. `The Design and Implementation of the 4.4 BSD
 * Operating System,' Addison-Wesley 1996, section 4.8 Process Groups
 * and Sessions.
 *
 ***************************************************************************/

#include "kernel_kernel.h"
#include "kernel_lgid.h"
#include "kernel_lpid.h"
#include "kernel_lprocess.h"

#include "lomac_plm.h"


#ifdef PARANOID
#define PLM_HASHNODE_MAGIC  0xAddedFee
#endif
#define PLM_HASHNUM         0x0000007f
#define PLM_HASH( pid )     ( (pid) & PLM_HASHNUM )


typedef struct plm_hashnode {
#ifdef PARANOID
  unsigned int magic;           /* magic number for paranoid assertions */
#endif
  lpid_t  pid;                  /* process pid */
  level_t level;                /* level associated with pid */
  trust_t trust;                /* trust setting for this pid */
  struct plm_hashnode *p_next;  /* points to next hashnode in chain */
} plm_hashnode_t;

plm_hashnode_t *plm_hash_table[ PLM_HASHNUM + 1 ];


/* ------------------------- Local Functions -------------------------- */

#ifdef PARANOID

int is_malformed_plm_hashnode( plm_hashnode_t * );

int is_malformed_plm_hashnode( plm_hashnode_t *p_hashnode ) {

  if( p_hashnode->magic != PLM_HASHNODE_MAGIC ) {
    return( 1 );
  }

  if( ( p_hashnode->level < LOMAC_LOWEST_LEVEL ) ||
      ( p_hashnode->level > LOMAC_HIGHEST_LEVEL ) ) {
    return( 1 );
  }

  if( ( p_hashnode->trust != e_untrusted ) &&
      ( p_hashnode->trust != e_neverdemote ) ) {
    return( 1 );
  }

  return( 0 );

} /* is_malformed_plm_hashnode() */

#endif

plm_hashnode_t *pid_to_plm_hashnode( lpid_t );


/* pid_to_plm_hashnode()
 *
 * in:     pid - pid to search for in the PLM
 * out:    nothing
 * return: a pointer to the PLM hashnode corresponding to `pid', or
 *         NULL if no such hashnode exists.
 *
 * This utility function is used to lookup hashnodes in the PLM.
 *
 */

plm_hashnode_t *pid_to_plm_hashnode( lpid_t pid ) {

  plm_hashnode_t *p_hashnode;  /* points to hashnodes in the PLM hash table */

  p_hashnode = plm_hash_table[ PLM_HASH( pid ) ];
  while( p_hashnode ) {

#ifdef PARANOID
    if( is_malformed_plm_hashnode( p_hashnode ) ) {
      panic( "Lomac (lomac_plm.c) : Bad PLM hashnode found on PLM query" );
    }
#endif
      
    if( p_hashnode->pid == pid ) {
      return( p_hashnode );
    }

    p_hashnode = p_hashnode->p_next;

  } /* for all hashnodes in chain */

  /* If we reach here, then `pid' does not exist in the PLM */
  return( NULL );

} /* pid_to_plm_hashnode() */



/* ------------------------- Global Functions -------------------------- */


/* plm_get()
 *
 * in:     pgid    - pgid of process group we wish to know about
 * out:    p_level - level of process group `pgid', or LOMAC_INVALID_LEVEL
 *         p_trust - trust setting of process group `pgid' or e_invalid
 * return: value     condition
 *         -----     ---------
 *           0       entry found for `pgid'.
 *           1       no entry found for `pgid'.
 *
 *  This function consults the PLM to determine the current level and
 *  trust settings for the process group indicated by `pgid'.
 *
 */

int plm_get( lpid_t pgid, level_t *p_level, trust_t *p_trust ) {

  plm_hashnode_t *p_hashnode;  /* points to PLM entry for `pgid' */

  if( !( p_hashnode = pid_to_plm_hashnode( pgid ) ) ) {

    /* The PLM does not contain an entry for `pgid'. */
    *p_level = LOMAC_INVALID_LEVEL;
    *p_trust = e_invalid;
    return( 1 );

  } /* if no entry for `pgid' in PLM */

  *p_level = p_hashnode->level;
  *p_trust = p_hashnode->trust;

#ifdef PARANOID
  if( is_malformed_plm_hashnode( p_hashnode ) ) {
    lpanic( "LOMAC(lomac_plm.c) - plm_get found malformed PLM hashnode" );
  }
#endif

  return( 0 );

} /* plm_get() */


/* plm_set()
 *
 * in:     pgid  - pgid of process group whose level we wish to set
 *         level - level to set `pgid's process group to
 *         trust - trust setting for `pgid'
 * out:    nothing
 * return: nothing
 *
 * Looks for a PLM entry corresponding to `pgid'.  If it does not find
 * one, it creates one.  Either way, it makes the entry's level equal
 * `level' and its trust flag equal `trust'.
 *
 */

void plm_set( lpid_t pgid, level_t level, trust_t trust ) {

  int hash_chain_index;        /* holds results of PLM_HASH() */
  plm_hashnode_t *p_hashnode;  /* points to hashnodes in the PLM hash table */

#ifdef PARANOID
  if( ( level < LOMAC_LOWEST_LEVEL ) || ( level > LOMAC_HIGHEST_LEVEL ) ) {
    lpanic( "LOMAC(lomac_plm.c) level out of bounds" );
  }
  if( ( trust != e_untrusted ) && ( trust != e_neverdemote ) ) {
    lpanic( "LOMAC(lomac_plm.c) bad trust setting" );
  }
#endif


  /* Make `p_hashnode' point to an existing entry for `pgid' or  *
   * to a newly-created entry if `pgid' didn't already have one. */
  if( !( p_hashnode = pid_to_plm_hashnode( pgid ) ) ) {
    
    /* If we reach here, then an entry for `pgid' does not exist in *
     * the PLM.  we must add a new hashnode to the PLM for `pgid'   *
     * at the head of the chain indicated by `hash_chain_index'.   */
    
    hash_chain_index = PLM_HASH( pgid );
    p_hashnode = (plm_hashnode_t *)lmalloc( sizeof( plm_hashnode_t ) );
    
#ifdef PARANOID
    if( !p_hashnode ) {
      lpanic( "Lomac (lomac_plm.c) : Unable to allocate new PLM hashnode" );
    }
    p_hashnode->magic = PLM_HASHNODE_MAGIC;
#endif
    
    p_hashnode->pid = pgid;
    p_hashnode->p_next = plm_hash_table[ hash_chain_index ];
    plm_hash_table[ hash_chain_index ] = p_hashnode;
    
  } /* if we had to create a new hashnode for `pgid' */

  /* Set the attributes we took in as parameters. */
  p_hashnode->level = level;
  p_hashnode->trust = trust;

#ifdef PARANOID
  if( !( p_hashnode = pid_to_plm_hashnode( pgid ) ) ) {
    lpanic( "LOMAC(lomac_plm.c) failed to add pgid to PLM" );
  }
  if( is_malformed_plm_hashnode( p_hashnode ) ) {
    lpanic( "LOMAC(lomac_plm.c) created bad PLM hashnode" );
  }
  if( level != p_hashnode->level ) {
    lpanic( "LOMAC(lomac_plm.c) failed to set PLM entry level" );
  }
  if( trust != p_hashnode->trust ) {
    lpanic( "LOMAC(lomac_plm.c) failed to set PLM entry trust" );
  }
#endif

} /* plm_set() */



/* init_plm()
 *
 * in:     nothing
 * out:    nothing
 * return: nothing
 *
 * Normally, we add process groups to the PLM when they are created by
 * the setpgid() or setsid() system calls.  However, when this LKM is
 * loaded, at least one process group will already be running.  We
 * must add all process groups which are running at the time this
 * module is loaded into the PLM.  Otherwise, when one of them calls a
 * read or write system call, we'll query the PLM looking for their
 * level, and not find it.  This function traverses the list of
 * processes and adds their process groups to the PLM at a default
 * high level.
 *
 * Note: we can't use the for_all_tasks() macro defined in sched.h to
 * traverse the population of processes because it relies on `init_task',
 * which is defined in sched.c and not exported to modules.  Although
 * it would be helpful if Linux exported this symbol, we can reimplement
 * the for_all_tasks() macro using current_set, which is exported.
 *
 */

void init_plm( void ) {

#ifdef PARANOID
  plm_hashnode_t *p_plm_entry;
#endif

  lprocess_t *p;   /* traverses list of existing processes */

  /* before we begin, clear `plm_hash_table' */
  memset( (void *)plm_hash_table, 0,
	  ( sizeof( plm_hashnode_t *) * PLM_HASHNUM + 1 ) );

  /* add all existing processes to PLM, grandfathering them *
   * to a high level.  Is this going to work on an SMP?     */
  p = lprocess_current();
  do {
    plm_set( lprocess_pgrp( p ), LOMAC_HIGHEST_LEVEL, e_untrusted );
    p = lprocess_next( p );
  } while( p != lprocess_current() );

#ifdef PARANOID
  p = lprocess_current();
  do {
    if( !( ( p_plm_entry = pid_to_plm_hashnode( lprocess_pgrp( p ) ) ) &&
           ( !is_malformed_plm_hashnode( p_plm_entry ) ) &&
           ( p_plm_entry->level == LOMAC_HIGHEST_LEVEL ) &&
           ( p_plm_entry->trust == e_untrusted ) ) ) {
      lpanic( "Lomac(lomac_plm.c) : failed to grandfather pid in PLM" );
    }
  } while( p != lprocess_current() );
#endif

} /* init_plm() */


/* update_plm_for_new_pgrp()
 *
 * in:     parent_pgrp - process group containing process that is creating
 *                       the new process group `child_pgrp'.
 *         child_pgrp  - the new process group.
 * out:    nothing
 * return: nothing
 *
 * This function adds an entry for `child_pgrp' into the PLM that has
 * the same level and trust attributes as `parent_pgrp'.  This
 * function causes new process groups to inherit their parent's level.
 * It must be invoked whenever a new process group is created, or
 * the PLM will become inconsistent with the state of the kernel.
 *
 */

void update_plm_for_new_pgrp( lpid_t parent_pgrp, lpid_t child_pgrp ) {

  level_t parent_level;
  trust_t parent_trust;

#ifdef PARANOID
  lprocess_t *p;   /* traverses list of existing processes */
  plm_hashnode_t *p_parent_entry;
  plm_hashnode_t *p_child_entry;

  if( !( p_parent_entry = pid_to_plm_hashnode( parent_pgrp ) ) ) {
    lpanic( "LOMAC(lomac_plm.c) - parent pgrp has no PLM entry" );
  }
  if( is_malformed_plm_hashnode( p_parent_entry ) ) {
    lpanic( "LOMAC(lomac_plm.c) - parent PLM entry corrupt" );
  }
#endif

  plm_get( parent_pgrp, &parent_level, &parent_trust );
  plm_set( child_pgrp, parent_level, parent_trust );

  /* lprint( KERN_INFO "pid %d pgrp %d spawns new pgrp %d at level %d\n",
   *         lprocess_pid( lprocess_current() ),
   * 	     parent_pgrp, child_pgrp, parent_level );
   */

#ifdef PARANOID
  if( !( p_child_entry = pid_to_plm_hashnode( child_pgrp ) ) ) {
    lpanic( "LOMAC(lomac_plm.c) - no PLM entry for new pgrp" );
  }
  if( is_malformed_plm_hashnode( p_child_entry ) ) {
    lpanic( "LOMAC(lomac_plm.c) - corrupt PLM entry for new pgrp" );
  }
  if( !( ( p_parent_entry->level == p_child_entry->level ) &&
         ( p_parent_entry->trust == p_child_entry->trust ) ) ) {
    lpanic( "LOMAC(lomac_plm.c) - bad inheritance for new pgrp" );
  }

  p = lprocess_current();
  do {
    if( !( pid_to_plm_hashnode( lprocess_pgrp( p ) ) ) ) {
      panic( "Lomac(lomac_plm.c) : PLM has new pgrp with no level" );
    }
  } while( p != lprocess_current() );
#endif

} /* update_plm_for_new_pgrp() */
  





