/****************************************************************************
 *
 * lomac_tpl.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 * This file contains the implementation of the Trusted Program List (TPL).
 *
 ****************************************************************************/

#include "kernel_kernel.h"
#include "kernel_lgid.h"
#include "kernel_lpid.h"
#include "kernel_lprocess.h"
#include "kernel_lvnode.h"


#include "lomac_tpl.h"
#include "lomac_plm.h"
#include "lomac_rta.h"

#ifdef PARANOID
#define TPL_LISTNODE_MAGIC   0xBeBadDad   /* magic number for assertions */
#endif

typedef struct tpl_listnode {
#ifdef PARANOID
  unsigned int magic;
#endif
  const char *filename_s;         /* filename of trusted program */
  struct tpl_listnode *p_next;    /* points to next node in list */
} tpl_listnode_t;

tpl_listnode_t *p_tpl_list_head;  /* head of TPL list */


/* ------------------------- Private Functions ------------------------ */

#ifdef PARANOID

int is_malformed_tpl_listnode( tpl_listnode_t * );

/* is_malformed_tpl_listnode()
 *
 * in:     p_node - node to check
 * out:    nothing
 * return: 1 if `p_node' is a malformed TPL list node, else 0.
 *
 */

int is_malformed_tpl_listnode( tpl_listnode_t *p_node ) {

  if( p_node->magic != TPL_LISTNODE_MAGIC ) {
    return( 1 );
  }

  return( is_malformed_absolute_path( p_node->filename_s ) );

} /* is_malformed_tpl_listnode() */

#endif

/* ------------------------- Public Functions ------------------------- */

/* init_tpl()
 *
 * in:     nothing
 * out:    nothing
 * return: nothing
 *
 * Initialize the TPL
 *
 */

void init_tpl( void ) {

  p_tpl_list_head = NULL;

  add_program_to_tpl( "/usr/sbin/syslogd" );
  /* add_program_to_tpl( "/usr/X11R6/bin/Xwrapper" );  */
  /* add_program_to_tpl( "/usr/X11R6/bin/XF86_SVGA" ); */

} /* init_tpl() */


/* add_program_to_tpl()
 *
 * in:     filename_s - absolute path, name of program to make trusted
 * out:    nothing
 * return: nothing
 *
 * Adds the program named by `filename_s' to the TPL.
 *
 */

void add_program_to_tpl( const char *filename_s ) {

  tpl_listnode_t *p_node;     /* used to traverse the TPL */

#ifdef PARANOID
  if( is_malformed_absolute_path( filename_s ) ) {
    lpanic( "LOMAC(lomac_tpl.c) bad path to trusted program" );
  }
#endif

  /* Search for an existing TPL entry for `filename_s'.  *
   * If found, return.                                   */
  for( p_node = p_tpl_list_head; p_node; p_node = p_node->p_next ) {

#ifdef PARANOID
    if( is_malformed_tpl_listnode( p_node ) ) {
      lpanic( "LOMAC(lomac_tpl.c) malformed node in TPL" );
    }
#endif

    if( ! strcmp( p_node->filename_s, filename_s ) ) {
      return;
    }

  } /* for all mappings in the TPL */

  /* If we arrive here, no mapping for `filename_s' exists *
   * in the TPL.  We must add one.                         */

  p_node = (tpl_listnode_t *)lmalloc( sizeof( tpl_listnode_t ) );

#ifdef PARANOID
  if( !p_node ) {
    lpanic( "Lomac(lomac_tpl.c) - unable to allocate new TPL listnode" );
  }
#endif

  p_node->filename_s = filename_s;
  p_node->p_next     = p_tpl_list_head;
  p_tpl_list_head    = p_node;

#ifdef PARANOID
  p_node->magic = TPL_LISTNODE_MAGIC;

  if( is_malformed_tpl_listnode( p_node ) ) {
    lpanic( "lomac(lomac_tpl.c) - created malformed TPL node" );
  }
  if( ! is_trusted_program( filename_s ) ) {
    lpanic( "lomac(lomac_elm.c) - failed to enter filename into TPL" );
  }
#endif

} /* add_program_to_tpl() */


/* is_trusted_program()
 * 
 * in:     filename_s - absolute path to program we wish to query in TPL.
 * out:    nothing
 * return: nothing
 *
 */

int is_trusted_program( const char *filename_s ) {

  tpl_listnode_t *p_node;      /* used to traverse TPL */
  lvnode_t *p_tpl_inode;       /* inode corresponding to TPL entry */
  lvnode_t *p_user_inode;      /* inode corresponding to `filename_s' parm */
  lvnode_id_t tpl_id;          /* unique id for `p_tpl_inode' */
  lvnode_id_t user_id;         /* unique id for `p_user_inode' */

  /* Make `p_user_inode' point to the inode corresponding to `filename_s'. */
  if( !( p_user_inode = lvnode_from_path( filename_s ) ) ) {
    return( 0 );  /* file does not exist, therefore it's not trusted */
  }
  lvnode_to_id( p_user_inode, &user_id );
  

  /* Traverse the TPL, translating each entry's filename into an inode *
   * pointed to by `p_tpl_inode'.  Look for matches to `p_user_inode'. */
  for( p_node = p_tpl_list_head; p_node; p_node = p_node->p_next ) {

#ifdef PARANOID
    if( is_malformed_tpl_listnode( p_node ) ) {
      lpanic( "LOMAC(lomac_tpl.c) - TPL query found bad list node" );
    }
#endif

    if( !( p_tpl_inode = lvnode_from_path( p_node->filename_s ) ) ) {
      /* This TPL entry's file doesn't exist.  Go to next entry */
      continue;
    }
    lvnode_to_id( p_tpl_inode, &tpl_id );

    if( lvnode_equals( &user_id, &tpl_id ) ) {
      /* we have a match - `filename_s' must be a trusted program */
      return( 1 );
    }
     
  } /* for all nodes in TPL list */

  /* `filename_s' is not in TPL, program is not trusted */
  return( 0 );

} /* is_trusted_program() */



