/*************************************************************************
 *
 * lomac_vlm.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 * This is the implementation of the Vnode Level Map (VLM).  In
 * general, the VLM contains the same information as the Exception
 * Level Map (ELM), but provides more efficient lookups.
 * Unfortunately, at any given time, the ELM is likely to contain
 * entries not found in the VLM, and the VLM is likely to contain
 * entries not found in the ELM.  This sorry state of affairs exists
 * because the relationship between pathnames and inodes is not a
 * function in the mathematical sense.
 *
 * 1.  Not every pathname corresponds to an inode.  Pathnames naming
 *     files which do not exist (have not been created, or have been
 *     deleted) do not have inodes.  The ELM may contain entries for
 *     nonexistent files, but because they have no inodes, the VLM
 *     cannot.
 * 2.  Not every inode corresponds to a pathname.  Inodes representing
 *     sockets, for example, have no pathname.  Processes can use the
 *     bind() system call to associate a local domain socket with a
 *     pathname, but the pathname itself is represented with its own
 *     inode, separate from the inode representing the socket.  The
 *     VLM may contain entries for inodes representing IPC objects like
 *     sockets, but since they lack pathnames, the ELM cannot.
 *
 *
 *
 * References:
 *
 * McKusick et. al. `The Design and Implementation of the 4.4 BSD
 * Operating System', Addison-Wesley, 1996, section 6.5: The
 * Virtual-Filesystem Interface.
 *
 * David A. Rusling `The Linux Kernel,' March 1998. (A description of
 * the linux kernel, available on the web,) section 9.2: The Virtual
 * File System (VFS).
 *
 *************************************************************************/

#include "kernel_kernel.h"
#include "kernel_lgid.h"
#include "kernel_lpid.h"
#include "kernel_lvnode.h"

#include "lomac_vlm.h"

#ifdef PARANOID
#define VLM_HASHNODE_MAGIC        0x2BadDeed
#endif
#define VLM_HASHNUM               0x1f
#define VLM_HASH( inode_number )  ( (inode_number) & VLM_HASHNUM )

typedef struct vlm_hashnode {
#ifdef PARANOID
  unsigned int magic;            /* magic number for paranoid assertions */
#endif
  lvnode_id_t     lvnode_id;     /* unique identifier for lvnode */
  level_t         level;         /* assigned level */
  struct vlm_hashnode *p_next;   /* next bucket in hash chain */
} vlm_hashnode_t;

vlm_hashnode_t *vlm_hash_table[ VLM_HASHNUM + 1 ];


/* ------------------------- Private Functions ------------------------ */


#ifdef PARANOID

int is_malformed_vlm_hashnode( vlm_hashnode_t * );

/* is_malformed_vlm_hashnode()
 *
 * in:     p_node - possibly bad pointer to potentially bad VLM node 
 * out:    nothig
 * return: 1 if `p_node' is malformed, else 0.
 *
 */

int is_malformed_vlm_hashnode( vlm_hashnode_t *p_node ) {

  if( !p_node ) {
    return( 1 );
  }

  if( p_node->magic != VLM_HASHNODE_MAGIC ) {
    return( 1 );
  }

  if( ( p_node->level < LOMAC_LOWEST_LEVEL ) ||
      ( p_node->level > LOMAC_HIGHEST_LEVEL ) ) {
    return( 1 );
  }
  

  return( 0 );

} /* is_malformed_vlm_hashnode() */

#endif PARANOID




/* -------------------------- Public Functions ------------------------ */

/* init_vlm()
 *
 * in:     nothing
 * out:    nothing
 * return: nothing
 *
 *
 */

void init_vlm( void ) {

  /* clear `vlm_hash_table' */
  memset( (void *)vlm_hash_table, 0,
	  ( sizeof( vlm_hashnode_t *) * VLM_HASHNUM + 1 ) );

} /* init_vlm() */


/* vlm_set()
 *
 * in:     p_lvnode_id - unique id of lvnode whose level we wish to set
 *         level       - level we want to assign to `lvnode_id'
 * out:    nothing
 * return: nothing
 *
 *
 */

void vlm_set( lvnode_id_t *p_lvnode_id, level_t level ) {

  vlm_hashnode_t *p_hashnode; /* points to hashnodes in the VLM hash table */
  int hash_chain_index;       /* holds results of VLM_HASH() */


#ifdef PARANOID
  if( ( level < LOMAC_LOWEST_LEVEL ) || ( level > LOMAC_HIGHEST_LEVEL ) ) {
    panic( "LOMAC(lomac_vlm.c) - bad level passed to vlm_set()" );
  }
#endif

  hash_chain_index = VLM_HASH( p_lvnode_id->lvnode_number );

  /* Make `p_hashnode' point to the hashnode that we want to make      *
   * correspond to `p_lvnode_id'.  We should first see if there is an  *
   * existing entry in the VLM for `p_lvnode_id', and if there is,     *
   * make `p_hashnode' point to it.  If there is no existing entry, we *
   * should make `p_hashnode' point to a freshly allocated and linked  *
   * one.                                                              */

  for( p_hashnode = vlm_hash_table[ hash_chain_index ];
       p_hashnode; p_hashnode = p_hashnode->p_next ) {

    if( lvnode_equals( &(p_hashnode->lvnode_id), p_lvnode_id ) ) {

      /* We have found an existing entry for `p_lvnode_id'.  Reuse it. */
      break;

    }

  } /* for all nodes in hash chain */

  if( !p_hashnode ) {

    /* We didn't find an existing entry to reuse, *
     * so we'll have to create a new one.         */

    p_hashnode = (vlm_hashnode_t *)lmalloc( sizeof( vlm_hashnode_t ) );

#ifdef PARANOID
    if( !p_hashnode ) {
      panic( "Lomac (lomac_vlm.c) : Unable to allocate new VLM hashnode" );
    }
    p_hashnode->magic = VLM_HASHNODE_MAGIC;
#endif

    p_hashnode->p_next                  = vlm_hash_table[ hash_chain_index ];
    vlm_hash_table[ hash_chain_index ]  = p_hashnode;
    p_hashnode->lvnode_id.lvnode_number = p_lvnode_id->lvnode_number;
    p_hashnode->lvnode_id.device_number = p_lvnode_id->device_number;

  } /* if we had to allocate a new node */

  p_hashnode->level                    = level;

#ifdef PARANOID
  if( is_malformed_vlm_hashnode( p_hashnode ) ) {
    panic( "Lomac (lomac_vlm.c) : Updated bad node in VLM" );
  }
  if( level != vlm_get( p_lvnode_id ) ) {
    panic( "LOMAC(lomac_vlm.c) failed to set lvnode level" );
  }
#endif

} /* vlm_set() */



/* vlm_get()
 *
 * in:     p_lvnode_id - id of lvnode whose level we wish to know.
 * out:    nothing
 * return: level of `p_lvnode_id' or 0 if no mapping exists.
 * 
 * Search the VLM for a mapping from `p_lvnode_id' to a level.  If we
 * find a mapping, return the indicated level.  If we find no mapping,
 * we return 0.
 * 
 */

level_t vlm_get( lvnode_id_t *p_lvnode_id ) {

  vlm_hashnode_t *p_hashnode; /* points to hashnodes in the VLM */

  p_hashnode = vlm_hash_table[ VLM_HASH( p_lvnode_id->lvnode_number ) ];
  while( p_hashnode ) {

#ifdef PARANOID
    if( is_malformed_vlm_hashnode( p_hashnode ) ) {
      panic( "Lomac(lomac_vlm.c) : Bad VLM hashnode" );
    }
#endif
      
    if( lvnode_equals( &(p_hashnode->lvnode_id), p_lvnode_id ) ) {

      return( p_hashnode->level );

    } /* if this hashnode matches the inode we're seeking */

    p_hashnode = p_hashnode->p_next;

  } /* for all hashnodes in chain */

  /* If we reach here, then `p_inode' does not exist in the VLM. */
  return( 0 );

} /* vlm_get() */


/* vlm_remove()
 *
 * in:     p_lvnode_id - id of vnode whose VLM entry we want to remove
 * out:    nothing
 * return: nothing
 *
 * Removes the entry corresponding to `p_lvnode_id' (if any) from the VLM.
 *
 */

void vlm_remove( lvnode_id_t *p_lvnode_id ) {

  vlm_hashnode_t **pp_parents_next; /* points to VLM hashnode next pointers */
  vlm_hashnode_t *p_remove_me;      /* points to the hashnode to remove */
  int hash_chain_index;             /* holds results of VLM_HASH() */

  hash_chain_index = VLM_HASH( p_lvnode_id->lvnode_number );

  /* Make `pp_parents_next' point to the `p_next' pointers of *
   * each of the hashnodes in the appropriate hash chain.     */
  for( pp_parents_next = &(vlm_hash_table[ hash_chain_index ]);
       *pp_parents_next; pp_parents_next = &((*pp_parents_next)->p_next) ) {

    if( lvnode_equals( &((*pp_parents_next)->lvnode_id), p_lvnode_id ) ) {

      /* pp_parents_next points to the next pointer which contains the *
       * address of the VLM hashnode we want to remove.  Save the      *
       * address of the hashnode we want to remove in `p_remove_me',   *
       * patch the next pointer to skip the hashnode we want to        *
       * remove, and then free the hashnode's memory.                  */
      p_remove_me = *pp_parents_next;
      *pp_parents_next = p_remove_me->p_next;
#ifdef PARANOID
      p_remove_me->magic = 0x00;  
#endif
      lfree( p_remove_me );

#ifdef PARANOID
      if( vlm_get( p_lvnode_id ) ) {
	lpanic( "LOMAC(lomac_vlm.c) - failed to remove VLM hashnode" );
      }
#endif

      return;

    } /* if we found the hashnode we want to remove */

  } /* for all nodes in hash chain */
  


} /* vlm_remove() */
